;; Generated package description from markchars.el  -*- no-byte-compile: t -*-
(define-package "markchars" "0.2.2" "Mark chars fitting certain characteristics" 'nil :commit "f069a7637a97239ecab7f34fd07fddfafb02ba3f" :url "https://elpa.gnu.org/packages/markchars.html" :authors '(("Lennart Borgman" . "lennart.borgman@gmail.com")) :maintainer '("Lennart Borgman" . "lennart.borgman@gmail.com"))
